using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;

// FluidEnviroEditorShower v1.0
//
// Allows you to edit some parameters of fluid particles

namespace FluidEnviroEditorShowerPlugin
{
    public class FluidEnviroEditorShower : MVRScript
    {
		
		public JSONStorableStringChooser FluidAppareance;
		protected List<string> appearanceChoices;
		public JSONStorableBool FluidStartOn;
		
		public JSONStorableColor FluidMainColor;
		public JSONStorableFloat FluidMainOpacity;
					
		public JSONStorableFloat StreamAmount;
		public JSONStorableFloat StreamSpeed;
		public JSONStorableFloat StreamForce;
		public JSONStorableFloat StreamAngle;
		public JSONStorableFloat StreamRadius;
		public JSONStorableFloat StreamThickness;
		public JSONStorableFloat StreamLifetimeMin;
		public JSONStorableFloat StreamLifetimeMax;
		
		public JSONStorableFloat FluidThickness;
		public JSONStorableFloat FluidStickyness;
		
		public JSONStorableFloat FluidLifetimeMin;
		public JSONStorableFloat FluidLifetimeMax;

		public Transform ParticleRoot;
		private ParticleSystem prtComponent;
		private ParticleSystemRenderer prtRendererComponent;
		
		private ParticleSystem prtComponentSub;
		private ParticleSystemRenderer prtRendererComponentSub;
		
		private ParticleSystem prtDeathComponent;
		private ParticleSystemRenderer prtDeathRendererComponent;
		
		private string objectSearchName = "hzm_fluidenviro_";

		private bool seInitialized = false;
		
		// Security
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		
		// Others
		private bool isSprayActive = false;
		
		// Curves
		private AnimationCurve cuGravPumping;
		private AnimationCurve cuGravContinuous;

		// Original values
		private float initialCollisionScale = 0f;
		
		// Actions
		public JSONStorableFloat A_SprayMultipleTimes;
		public JSONStorableString A_SprayMultipleTimesCustom;
		public JSONStorableAction A_StartStream;
		public JSONStorableAction A_StopStream;
		
		public JSONStorableColor A_ChangeFluidMainColor;
		public JSONStorableFloat A_ChangeFluidMainOpacity;
		public JSONStorableFloat A_ChangeStreamAmount;
		public JSONStorableFloat A_ChangeStreamSpeed;
		public JSONStorableFloat A_ChangeStreamForce;
		public JSONStorableFloat A_ChangeStreamAngle;
		public JSONStorableFloat A_ChangeStreamRadius;
		public JSONStorableFloat A_ChangeStreamThickness;
		public JSONStorableFloat A_ChangeFluidThickness;
		public JSONStorableFloat A_ChangeFluidStickyness;
		public JSONStorableAction A_ResetAppearanceToDefault;
		
		public override void Init()
        {			 
            try
            {
				if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add FluidEnviroEditorShower on an Custom Unity Asset with a compatible particle system.");
                    return;
                }

				// Type of appearance (lighting)
				appearanceChoices = new List<string>();
				appearanceChoices.Add("Unlit");
				appearanceChoices.Add("Lit");
							
				FluidAppareance = new JSONStorableStringChooser("Fluid appearance", appearanceChoices, "Unlit", "Fluid appearance", FluidAppareanceCallback);
				UIDynamicPopup FAdp = CreatePopup(FluidAppareance, false);
				FAdp.labelWidth = 300f;
				
				FluidStartOn = new JSONStorableBool("Fluid start on", true, FluidStartOnCallback);
				FluidStartOn.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(FluidStartOn, false);
				
				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				FluidMainColor = new JSONStorableColor("Fluid color", hsvc, MainColorCallback);
				CreateColorPicker(FluidMainColor);
				
				FluidMainOpacity = new JSONStorableFloat("Main fluid opacity", 0.15f, MainOpacityCallback, 0f, 1f);
				CreateSlider(FluidMainOpacity, false);

				StreamAmount = new JSONStorableFloat("Stream amount", 350f, StreamAmountCallback, 10f, 2000f);
				CreateSlider(StreamAmount, false);
				
				StreamSpeed = new JSONStorableFloat("Stream speed", 1f, StreamSpeedCallback, 0.1f, 1f);
				CreateSlider(StreamSpeed, false);
				
				StreamForce = new JSONStorableFloat("Stream force", 3.0f, StreamForceCallback, 0f, 6f);
				CreateSlider(StreamForce, false);
				
				StreamAngle = new JSONStorableFloat("Stream angle", 6.0f, StreamAngleCallback, 0.5f, 45f);
				CreateSlider(StreamAngle, false);
				
				StreamRadius = new JSONStorableFloat("Stream radius", 0.4f, StreamRadiusCallback, 0.1f, 2f);
				CreateSlider(StreamRadius, false);
				
				StreamThickness = new JSONStorableFloat("Stream thickness", 0.8f, StreamThicknessCallback, 0.2f, 5f);
				CreateSlider(StreamThickness, false);

				StreamLifetimeMin = new JSONStorableFloat("Stream lifetime (min)", 1.5f, StreamLifetimeMinCallback, 0.5f, 5f);
				CreateSlider(StreamLifetimeMin, false);
				
				StreamLifetimeMax = new JSONStorableFloat("Stream lifetime (max)", 1.5f, StreamLifetimeMaxCallback, 0.5f, 5f);
				CreateSlider(StreamLifetimeMax, false);
				
				FluidThickness = new JSONStorableFloat("Fluid thickness", 1.0f, FluidThicknessCallback, 0.2f, 5f);
				CreateSlider(FluidThickness, false);
				
				FluidStickyness = new JSONStorableFloat("Fluid stickiness", 0.1f, FluidStickynessCallback, 0.01f, 0.99f);
				CreateSlider(FluidStickyness, false);
				
				FluidLifetimeMin = new JSONStorableFloat("Fluid lifetime (min)", 1.2f, FluidLifetimeMinCallback, 0.5f, 5f);
				CreateSlider(FluidLifetimeMin, false);
				
				FluidLifetimeMax = new JSONStorableFloat("Fluid lifetime (max)", 1.3f, FluidLifetimeMaxCallback, 0.5f, 5f);
				CreateSlider(FluidLifetimeMax, false);
							
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help - Shower</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>IMPORTANT:</b> If you switch particles in the Asset tab of the CUA, don't forget to reload the script to enable customization.\n\n" +
					"<b>NOTES:</b> This is an expensive physical particle system. Default parameters are meant to run properly. If you use continuous streams, performance can be impacted. For VR / Gameplay usage, 350 to 450 in the amount is a good ratio. High values are meant to be used by screenshot enthousiasts.\n\n" +
					"<b>Fluid appearance</b> is the way fluid looks depending on the light.\n" +
					"<b><i>- Unlit</i></b> ignores the light which means the fluid will be visible in complete darkness.\n" +
					"<b><i>- Lit</i></b> takes in account the lighting, which means the fluid will be invisible in complete darkness but it will also be coloured by the lights around.\n\n" +
					"<b>FUNCTIONS</b>\nYou have several function available to change the appearance during the scene.\n\n" +
					"All these functions starts with <b>Change xxxx</b>.\n\nThey will allow you to change the appearance of the fluid during the scene without changing the default setup of your fluid.\n\n" +
					"There is also a function allowing you to reset the appearance to the default one configured in the plugin called <b>Reset settings to default</b>\n\n" +
					"<b>TRIGGERS</b>\nYou have two triggers available\n\n" +
					"<b>- Start stream</b>\n" +
					"  Start shooting and never stops.\n\n" +
					"<b>- Stop stream</b>\n" +
					"  Stop shooting.\n\n" +
					"</color>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1080.0f;
							
				RegisterStringChooser(FluidAppareance);
				RegisterBool(FluidStartOn);
				RegisterColor(FluidMainColor);
				RegisterFloat(FluidMainOpacity);
				RegisterFloat(StreamAmount);
				RegisterFloat(StreamSpeed);
				RegisterFloat(StreamForce);
				RegisterFloat(StreamAngle);
				RegisterFloat(StreamRadius);
				RegisterFloat(StreamThickness);
				RegisterFloat(StreamLifetimeMin);
				RegisterFloat(StreamLifetimeMax);
				RegisterFloat(FluidThickness);
				RegisterFloat(FluidStickyness);
				RegisterFloat(FluidLifetimeMin);
				RegisterFloat(FluidLifetimeMax);
				
				// ************************************************
				// ** ACTIONS
				// ************************************************
				A_SprayMultipleTimes = new JSONStorableFloat("Spray X times, random duration", 0f, A_SprayMultipleTimesCallback, 0f, 20f){ isStorable = false };
				A_SprayMultipleTimesCustom = new JSONStorableString("Spray X times, customizable", "", A_SprayMultipleTimesCustomCallback){ isStorable = false };
				
				A_StartStream = new JSONStorableAction("Start stream", () => { StartStream(); });
				A_StopStream = new JSONStorableAction("Stop stream", () => { StopStream(); });
				
				HSVColor hsvc2 = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				A_ChangeFluidMainColor = new JSONStorableColor("Change fluid color", hsvc2, A_MainColorCallback){ isStorable = false };
				
				A_ChangeFluidMainOpacity = new JSONStorableFloat("Change fluid opacity", 0f, A_MainOpacityCallback, 0f, 1f){ isStorable = false };
				A_ChangeStreamAmount = new JSONStorableFloat("Change stream amount", 0f, StreamAmountCallback, 10f, 500f){ isStorable = false };
				A_ChangeStreamSpeed = new JSONStorableFloat("Change stream speed", 0f, StreamSpeedCallback, 0.1f, 1f){ isStorable = false };
				A_ChangeStreamForce = new JSONStorableFloat("Change stream force", 0f, StreamForceCallback, 0f, 6f){ isStorable = false };
				A_ChangeStreamAngle = new JSONStorableFloat("Change stream angle", 0f, StreamAngleCallback, 0.5f, 45f){ isStorable = false };
				A_ChangeStreamRadius = new JSONStorableFloat("Change stream radius", 0f, StreamRadiusCallback, 0.1f, 2f){ isStorable = false };
				A_ChangeStreamThickness = new JSONStorableFloat("Change stream thickness", 0f, StreamThicknessCallback, 0.2f, 5f){ isStorable = false };
				A_ChangeFluidThickness = new JSONStorableFloat("Change fluid thickness", 0f, FluidThicknessCallback, 0.2f, 5f){ isStorable = false };
				A_ChangeFluidStickyness = new JSONStorableFloat("Change fluid stickyness", 0f, FluidStickynessCallback, 0.2f, 0.99f){ isStorable = false };
				
				A_ResetAppearanceToDefault = new JSONStorableAction("Reset settings to default", () => { UpdateParticleSettings(); });
				
				RegisterFloat(A_SprayMultipleTimes);
				RegisterString(A_SprayMultipleTimesCustom);
				RegisterAction(A_StartStream);
				RegisterAction(A_StopStream);
				
				RegisterColor(A_ChangeFluidMainColor);
				RegisterFloat(A_ChangeFluidMainOpacity);
				RegisterFloat(A_ChangeStreamAmount);
				RegisterFloat(A_ChangeStreamSpeed);
				RegisterFloat(A_ChangeStreamForce);
				RegisterFloat(A_ChangeStreamAngle);
				RegisterFloat(A_ChangeStreamRadius);
				RegisterFloat(A_ChangeStreamThickness);
				RegisterFloat(A_ChangeFluidThickness);
				RegisterFloat(A_ChangeFluidStickyness);
				RegisterAction(A_ResetAppearanceToDefault);

			}
            catch(Exception e)
            {
                SuperController.LogError("FluidEnviroEditorShower - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }
		
		void Update()
        {
			// If we haven't made over 1min of init tests, we aren't initialized and the particle system (CUA) is still compatible, try to initialize
			if( initLoops < maxInitLoops && seInitialized == false ) {
				initLoops++;
				InitDE();
			}
			
			if( initLoops == maxInitLoops ) {
				logError("FluidEnviroEditorShower : This CUA particle could not be initialized properly after 3600 tries. Please select a compatible particle and reload the script.");
				initLoops++; // To avoid spamming the log
			}
		}
		
		private bool InitDE() {
			if( seInitialized == true ) return true;
			
			ParticleRoot = getObjectRoot( containingAtom.reParentObject );
			if( ParticleRoot == null ) {
				return false;
			}

			foreach( Transform child in ParticleRoot )
			{
				prtComponent = child.GetComponent<ParticleSystem>();
				if( prtComponent ) {
					prtRendererComponent = child.GetComponent<ParticleSystemRenderer>();
					
					foreach( Transform subChild in child )
					{
						if( subChild.name == "fluid_on_death" ) {
							prtDeathComponent = subChild.GetComponent<ParticleSystem>();
							if( prtDeathComponent ) {
								prtDeathRendererComponent = subChild.GetComponent<ParticleSystemRenderer>();
								
								var prtDeathCollision = prtDeathComponent.collision;
								initialCollisionScale = prtDeathCollision.radiusScale;
							}
						}
					}

					if( !prtDeathComponent || !prtDeathRendererComponent ) return false;
					seInitialized = true;
					
					UpdateParticleSettings(); // Sets all the parameters of the particle
					ChangeFluidChildParticle( FluidAppareance.val ); // Change the particle appearance
										
					logDebug("FluidEnviroEditorShower : Fluid editor initialized properly");

					return true;
				} else {
					logError("FluidEnviroEditorShower : your CUA is incorrect and cannot be used with this script.");
					initLoops = maxInitLoops; // We don't have a particle system, we consider the initialization failed
					return false;
				}
				
				break;
			}
			
			return false;

		}
		
		private Transform getObjectRoot( Transform parent ) {
			foreach( Transform child in parent )
			{
				if( child.name.Contains(objectSearchName) ) {
					return child;
				} else {
					Transform childSearch = getObjectRoot(child);
					if (childSearch != null) return childSearch;
				}
			}
			
			return null;
		}
		
		private void UpdateParticleSettings() {
			FluidStartOnCallback( FluidStartOn.val );
			
			MainColorCallback( FluidMainColor );
			MainOpacityCallback( FluidMainOpacity.val );
			
			StreamAmountCallback( StreamAmount.val );
			StreamSpeedCallback( StreamSpeed.val );
			StreamForceCallback( StreamForce.val );
			StreamAngleCallback( StreamAngle.val );
			StreamRadiusCallback( StreamRadius.val );
			
			StreamThicknessCallback( StreamThickness.val );
			FluidThicknessCallback( FluidThickness.val );
			
			FluidStickynessCallback ( FluidStickyness.val );
			
			FluidLifetimeMinCallback( FluidLifetimeMin.val );
		}
		
		protected void ChangeFluidChildParticle( string particleName ) {
			if( ParticleRoot != null ) {
				int loopCount = 0;
				foreach( Transform child in ParticleRoot )
				{
					if( loopCount == appearanceChoices.IndexOf(particleName) ) {
						child.gameObject.SetActive(true);
						prtComponent = child.GetComponent<ParticleSystem>();
						if( prtComponent ) {
							prtRendererComponent = child.GetComponent<ParticleSystemRenderer>();
							
							foreach( Transform subChild in child )
							{
								// Looking for first child with this name
								if( subChild.name == "fluid_on_death" ) {
									prtDeathComponent = subChild.GetComponent<ParticleSystem>();
									if( prtDeathComponent ) {
										prtDeathRendererComponent = subChild.GetComponent<ParticleSystemRenderer>();
										
										var prtDeathCollision = prtDeathComponent.collision;
										initialCollisionScale = prtDeathCollision.radiusScale;
									}									
								}
								
								break;
							}
						}
					} else {
						child.gameObject.SetActive(false);
					}
					
					loopCount++;
				}
				
				// Setting parameters
				UpdateParticleSettings();
			}
		}
		
		protected void StartStream() {
			prtComponent.Play();
		}
		
		protected void StopStream() {
			prtComponent.Stop();
		}

		// **************************
		// Callbacks & similar
		// **************************
		protected void FluidAppareanceCallback(string appearanceChoice) {
			ChangeFluidChildParticle( appearanceChoice );
		}
		
		protected void FluidStartOnCallback(bool enabled) {
			if( prtComponent ) {
				if( enabled == true ) {
					StartStream();
				} else {
					StopStream();
				}
			}
		}
				
		protected void MainColorCallback(JSONStorableColor color) {
			if( prtComponent ) {
				Color mainColor = new Color(color.colorPicker.currentColor.r, color.colorPicker.currentColor.g, color.colorPicker.currentColor.b, FluidMainOpacity.val);
				
				var prtMain = prtComponent.main;
				prtMain.startColor = mainColor;
			}
		}
		
		protected void MainOpacityCallback(float opacity) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startColor = new Color(FluidMainColor.colorPicker.currentColor.r, FluidMainColor.colorPicker.currentColor.g, FluidMainColor.colorPicker.currentColor.b, opacity);
			}
		}
				
		protected void StreamAmountCallback(float amount) {
			if( prtComponent ) {
				var prtEmission = prtComponent.emission;
				prtEmission.rateOverTime = amount;
			}
		}
		
		protected void StreamSpeedCallback(float speed) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.simulationSpeed = speed;
			}
		}
		
		protected void StreamForceCallback(float force) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startSpeed = force;
			}
		}
		
		protected void StreamAngleCallback(float angle) {
			if( prtComponent ) {
				var prtShape = prtComponent.shape;
				prtShape.angle = angle;
			}
		}
		
		protected void StreamRadiusCallback(float radius) {
			if( prtComponent ) {
				var prtShape = prtComponent.shape;
				prtShape.radius = radius;
			}
		}
		
		protected void StreamThicknessCallback(float thickness) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startSize = thickness;
			}
		}
		
		protected void StreamLifetimeMinCallback(float lifetimeMin) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startLifetime = new ParticleSystem.MinMaxCurve(lifetimeMin, StreamLifetimeMax.val);
			}
		}
		
		protected void StreamLifetimeMaxCallback(float lifetimeMax) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startLifetime = new ParticleSystem.MinMaxCurve(StreamLifetimeMin.val, lifetimeMax);
			}
		}
		
		protected void FluidThicknessCallback(float thickness) {
			if( prtComponent ) {			
				var prtDeathMain = prtDeathComponent.main;
				prtDeathMain.startSize = thickness;
				
				var prtDeathCollision = prtDeathComponent.collision;
				prtDeathCollision.radiusScale = initialCollisionScale * thickness;
			}
		}
		
		protected void FluidStickynessCallback(float stickyness) {
			if( prtComponent ) {
				var prtDeathCollision = prtDeathComponent.collision;
				prtDeathCollision.dampen = stickyness;
			}
		}
		
		protected void FluidLifetimeMinCallback(float lifetimeMin) {
			if( prtComponent ) {
				var prtDeathMain = prtDeathComponent.main;
				prtDeathMain.startLifetime = new ParticleSystem.MinMaxCurve(lifetimeMin, FluidLifetimeMax.val);
			}
		}
		
		protected void FluidLifetimeMaxCallback(float lifetimeMax) {
			if( prtComponent ) {
				var prtDeathMain = prtDeathComponent.main;
				prtDeathMain.startLifetime = new ParticleSystem.MinMaxCurve(FluidLifetimeMin.val, lifetimeMax);
			}
		}
			
		private void A_SprayMultipleTimesCallback(float numtimes) {
			DoSprayMultipleTimesRandomDuration( (int)Mathf.Round(numtimes) );
			A_SprayMultipleTimes.val = 0f;
		}
		
		private void A_SprayMultipleTimesCustomCallback(string customText) {
			DoSprayMultipleTimesCustom( customText );
			A_SprayMultipleTimesCustom.val = "";
		}
		
		protected void A_MainColorCallback(JSONStorableColor color) {
			if( prtComponent ) {
		
				var selectedColor = HSVColorPicker.HSVToRGB(color.val);
				
				var prtMain = prtComponent.main;
				Color newColor = new Color(selectedColor.r, selectedColor.g, selectedColor.b, prtMain.startColor.color.a);
				prtMain.startColor = newColor;
			}
		}
		
		protected void A_MainOpacityCallback(float opacity) {
			if( prtComponent ) {
				var prtMain = prtComponent.main;
				prtMain.startColor = new Color(prtMain.startColor.color.r, prtMain.startColor.color.g, prtMain.startColor.color.b, opacity);
			}
		}
		
		private void DoSprayMultipleTimesRandomDuration(int numtimes) {
			if( isSprayActive == false ) {
				StartCoroutine(SprayMultipleTimesRandomDuration(numtimes));
			}
		}
		
		private void DoSprayMultipleTimesCustom(string customText) {
			if( isSprayActive == false ) {
				StartCoroutine(SprayMultipleTimesCustom(customText));
			}
		}

		// **************************
		// COROUTINES
		// **************************
		private IEnumerator SprayMultipleTimesRandomDuration(int numtimes) 
		{
			// Spraying, preventing another to start
			isSprayActive = true;

			int sprayLoops = 0;
			while (sprayLoops < numtimes)
			{
				prtComponent.Play();
				
				float sprayTime = UnityEngine.Random.Range(0.1f,1f);
				yield return new WaitForSeconds(sprayTime);
				
				prtComponent.Stop();
				float pauseTime = UnityEngine.Random.Range(0.2f,0.4f);
				yield return new WaitForSeconds(pauseTime);
				
				sprayLoops++;
			}
			
			
			// Spray done, we can trigger another one
			isSprayActive = false;
		}
		
		private IEnumerator SprayMultipleTimesCustom(string customText) 
		{
			// Spraying, preventing another to start
			isSprayActive = true;

			// Splitting my custom text
			string[] seq = customText.Split(';');
			
			// Doing the loop !
			if( seq.Length > 0 ) {
				foreach (string seqElement in seq)
				{
					string[] sprayElem = seqElement.Split(',');
					if( sprayElem.Length == 2 ) {
						prtComponent.Play();
						yield return new WaitForSeconds(float.Parse(sprayElem[0]));
						
						prtComponent.Stop();
						yield return new WaitForSeconds(float.Parse(sprayElem[1]));
					}
				}
			}
			
			// Spray done, we can trigger another one
			isSprayActive = false;
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
	}
}